from .module import Module


class Datadog(Module):
    dependencies = ['task']

    def __init__(self, **kwargs):
        super().__init__('datadog', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='datadog help')

        p = subp.add_parser('mk', help='make a datadog task')
        p.add_argument('key', metavar='KEY', help='API key')
        p.set_defaults(datadog_handler=self.handle_make)

    def handle_make(self, args):
        self.make(args.key)

    def make(self, key):
        task_mod = self.get_module('task')
        task_mod.make('dd_agent', '!datadog/docker-dd-agent:latest', memory=4, logs=False)
        task_mod.env_set(
            'dd_agent',
            {
                'API_KEY': key,
                'SD_BACKEND': 'docker',
            }
        )
        task_mod.volume_add('dd_agent', 'socket', '/var/run/docker.sock', '/var/run/docker.sock', read_only=True)
        task_mod.volume_add('dd_agent', 'proc', '/host/proc/', '/proc/', read_only=True)
        task_mod.volume_add('dd_agent', 'cgroup', '/host/sys/fs/cgroup', '/cgroup/', read_only=True)



import pprint

from .module import Module


class Session(Module):
    def __init__(self, **kwargs):
        super().__init__('session', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='session help')

        p = subp.add_parser('sv', help='save session')
        p.add_argument('name', help='session name')
        p.set_defaults(session_handler=self.handle_save)

        p = subp.add_parser('ld', help='load session')
        p.add_argument('name', help='session name')
        p.set_defaults(session_handler=self.handle_load)

        p = subp.add_parser('ls', help='list sessions')
        p.set_defaults(session_handler=self.handle_list)

        p = subp.add_parser('sh', help='show current state')
        p.set_defaults(session_handler=self.handle_show)

    def handle_save(self, args):
        cache = {}
        for mod in self.client.modules:
            cache[mod.name] = mod.save()
        self.store[args.name] = cache

    def handle_load(self, args):
        cache = self.store.get(args.name, {})
        for mod in self.client.modules:
            mod.load(cache.get(mod.name, {}))

    def handle_list(self, args):
        for name in self.store.keys():
            print(name)

    def handle_show(self, args):
        for mod in self.client.modules:
            cache = mod.save()
            if cache:
                print('{}:'.format(mod.name))
                print('  {}'.format(pprint.pformat(cache, indent=3)))



import os
import json

from .module import Module
from .db import get_rc_path
from .utils import gen_secret, gen_name


class Pg(Module):
    dependencies = ['app']

    def __init__(self, **kwargs):
        super().__init__('pg', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='pg help')

        p = subp.add_parser('ls', help='list DBs')
        p.add_argument('name', metavar='NAME', nargs='?', help='DB name')
        p.set_defaults(pg_handler=self.handle_list)

        p = subp.add_parser('mk', help='add a postgres database')
        p.add_argument('name', metavar='NAME', help='DB name')
        p.add_argument('--backup', '-b', default=7, type=int, help='number of days to retain backups')
        p.set_defaults(pg_handler=self.handle_make)

        p = subp.add_parser('cache', help='cache DB details')
        p.add_argument('name', metavar='NAME', help='DB name')
        p.add_argument('password', metavar='PASSWORD', help='DB password')
        p.set_defaults(pg_handler=self.handle_cache)

        p = subp.add_parser('connect', help='connect to a task')
        p.add_argument('name', metavar='NAME', help='DB name')
        p.add_argument('target', metavar='TARGET', nargs='?', help='target task name')
        p.set_defaults(pg_handler=self.handle_connect)

        p = subp.add_parser('sl', help='select a postgres database')
        p.add_argument('name', metavar='NAME', nargs='?', help='database to select')
        p.add_argument('--show', '-s', action='store_true', help='show currently db')
        p.set_defaults(pg_handler=self.handle_select)

        p = subp.add_parser('psql')
        p.add_argument('name', metavar='NAME', help='DB name')
        p.set_defaults(pg_handler=self.handle_psql)

        # p = subp.add_parser('dump')
        # p.add_argument('output')
        # p.set_defaults(pg_handler=self.handle_dump)

        p = subp.add_parser('restore', help='restore a database')
        p.add_argument('input', metavar='INPUT', help='database dump file')
        p.set_defaults(pg_handler=self.handle_restore)

    def handle_list(self, args):
        self.list(args.name)

    def list(self, name):
        ctx = self.get_context()
        app = ctx['app']
        rds = self.get_boto_client('rds')
        if name:
            data = rds.describe_db_instances(
                Filter=[{
                    'Key': 'Name',
                    'Values': [self.get_id(name)]
                }]
            )
            print(json.dumps(data, indent=2))
        else:
            data = rds.describe_db_instances()
            pre = self.get_id('')
            for db in data['DBInstances']:
                name = db['DBInstanceIdentifier']
                if name.startswith(pre):
                    print(name[len(pre):])

    def handle_make(self, args):
        self.make(args.name, args.backup)

    def make(self, name, backup):
        ctx = self.get_context()
        app = ctx['app']
        password = gen_secret(16)
        inst_id = self.get_id(name)
        sg_id = self.client.get_module('cluster').get_security_group_id()
        rds = self.get_boto_client('rds')
        rds.create_db_subnet_group(
            DBSubnetGroupName=inst_id,
            DBSubnetGroupDescription=f'Subnet group for {inst_id}',
            SubnetIds=[sn.id for sn in self.get_module('cluster').iter_public_subnets(ctx['cluster'])]
        )
        data = rds.create_db_instance(
            DBName=name,
            DBInstanceIdentifier=inst_id,
            DBInstanceClass='db.t2.micro',
            Engine='postgres',
            AllocatedStorage=5,
            MasterUsername=name,
            MasterUserPassword=password,
            BackupRetentionPeriod=backup,
            DBSubnetGroupName=inst_id,
            VpcSecurityGroupIds=[sg_id],
            PubliclyAccessible=True,
            Tags=[
                {
                    'Key': 'cluster',
                    'Value': ctx['cluster']
                },
                {
                    'Key': 'app',
                    'Value': app
                }
            ]
        )
        waiter = rds.get_waiter('db_instance_available')
        waiter.wait(
            DBInstanceIdentifier=inst_id
        )
        self.cache(name, password)

    def handle_cache(self, args):
        self.cache(args.name, args.password)

    def cache(self, name, password):
        ctx = self.get_context()
        data = self.get_endpoint(name)
        path = os.path.join(self.get_rc_path(), f'{name}.pgpass')
        try:
            os.makedirs(os.path.dirname(path))
        except OSError:
            pass
        with open(path, 'w') as outf:
            outf.write('{}:{}:{}:{}:{}'.format(
                data['Address'],
                data['Port'],
                name,
                name,
                password
            ))
        os.chmod(path, 0o600)
        self.run(
            'gpg -c {}'.format(path)
        )
        s3 = self.get_boto_client('s3')
        s3.upload_file(f'{path}.gpg', ctx['bucket'], f'fuku/{ctx["cluster"]}/{ctx["app"]}/{name}.pgpass.gpg')

    def handle_connect(self, args):
        self.connect(args.name, args.target)

    def connect(self, db_name, task_name):
        task_mod = self.client.get_module('task')
        env = {
            'DATABASE_URL': self.get_url(db_name)
        }
        task_mod.env_set(task_name, env)

    def handle_select(self, args):
        self.select(args)

    def select(self, args):
        if args.show:
            sel = self.get_selected(fail=False)
            if sel:
                print(sel)
        else:
            name = args.name
            if name:
                # self.exists(name)
                self.get_pgpass_file(name)
                self.store['selected'] = name
            else:
                try:
                    del self.store['selected']
                except KeyError:
                    pass
            self.clear_parent_selections()

    def handle_psql(self, args):
        self.psql(args.name)

    def psql(self, name):
        ctx = self.get_context()
        path = os.path.join(self.get_rc_path(), f'{name}.pgpass')
        endpoint = self.get_endpoint(name)
        self.run(
            'psql -h {} -p {} -U {} -d {}'.format(
                endpoint['Address'],
                endpoint['Port'],
                name,
                name
            ),
            capture=False,
            env={'PGPASSFILE': path}
        )

    def handle_dump(self, args):
        ctx = self.get_context()
        app = self.client.get_selected('app')
        name = self.get_selected()
        path = os.path.join(self.get_rc_path(), '%s.pgpass' % name)
        endpoint = self.get_endpoint(name)
        self.run(
            'pg_dump -Fc --no-acl --no-owner -h {} -p {} -U {} -d {} -f {}'.format(
                endpoint['Address'],
                endpoint['Port'],
                name,
                name,
                args.output
            ),
            capture=False,
            env={'PGPASSFILE': path}
        )

    def handle_restore(self, args):
        ctx = self.get_context()
        name = self.get_selected()
        path = os.path.join(self.get_rc_path(), '%s.pgpass' % name)
        endpoint = self.get_endpoint(name)
        self.run(
            f'psql -h {endpoint["Address"]} -p {endpoint["Port"]} -U {name} {name} -c \'DROP SCHEMA public CASCADE; CREATE SCHEMA public;\'',
            env={'PGPASSFILE': path}
        )
        self.run(
            f'pg_restore --no-acl --no-owner -h {endpoint["Address"]} -p {endpoint["Port"]} -U {name} -d {name} {args.input}',
            capture=False,
            env={'PGPASSFILE': path}
        )

    def get_id(self, name):
        ctx = self.get_context()
        return f'fuku-{ctx["cluster"]}-{ctx["app"]}-{name}'

    def get_rc_path(self):
        ctx = self.get_context()
        return os.path.join(get_rc_path(), ctx['cluster'], ctx['app'])

    def get_endpoint(self, name):
        ctx = self.get_context()
        inst_id = self.get_id(name)
        rds = self.get_boto_client('rds')
        try:
            return rds.describe_db_instances(
                DBInstanceIdentifier=inst_id
            )['DBInstances'][0]['Endpoint']
        except:
            self.error(f'no database "{name}"')

    def get_url(self, name):
        ctx = self.get_context()
        path = os.path.join(self.get_rc_path(), f'{name}.pgpass')
        try:
            with open(path, 'r') as inf:
                data = inf.read()
        except:
            self.error(f'no cached information for "{name}"')
        host, port, db, user, pw = data.split(':')
        return 'postgres://{}:{}@{}:{}/{}'.format(user, pw, host, port, db)

    def get_pgpass_file(self, name):
        ctx = self.get_context()
        app = self.client.get_selected('app')
        path = os.path.join(self.get_rc_path(), '%s.pgpass' % name)
        if not os.path.exists(path):
            self.run(
                '$aws s3 cp s3://$bucket/fuku/{}/{}/{}.pgpass.gpg {}.gpg'.format(app, ctx['cluster'], name, path)
            )
            self.run(
                'gpg -o {} -d {}.gpg'.format(path, path)
            )
            os.chmod(path, 0o600)

    def get_selected(self, fail=True):
        sel = self.store.get('selected', None)
        if not sel and fail:
            self.error('no postgres DB selected')
        return sel



import json

from .module import Module
from .utils import (
    StoreKeyValuePair, StorePortPair,
    env_to_dict, dict_to_env,
    ports_to_dict, dict_to_ports,
    volumes_to_dict, dict_to_volumes,
    mounts_to_dict, dict_to_mounts,
    entity_already_exists
)


class Task(Module):
    dependencies = ['app']

    def __init__(self, **kwargs):
        super().__init__('task', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='task help')

        p = subp.add_parser('ls', help='list tasks')
        p.add_argument('name', metavar='NAME', nargs='?', help='task name')
        p.set_defaults(task_handler=self.handle_list)

        p = subp.add_parser('mk', help='make a task')
        p.add_argument('name', metavar='NAME', help='task name')
        p.add_argument('image', metavar='IMAGE', help='image name')
        p.add_argument('--cpu', '-c', help='cpu reservation')
        p.add_argument('--memory', '-m', help='memory reservation (MiB)')
        p.set_defaults(task_handler=self.handle_make)

        p = subp.add_parser('up', help='update a task')
        p.add_argument('name', metavar='NAME', help='task name')
        p.add_argument('image', metavar='IMAGE', help='image name')
        p.add_argument('--cpu', '-c', help='cpu reservation')
        p.add_argument('--memory', '-m', help='memory reservation (MiB)')
        p.set_defaults(task_handler=self.handle_update)

        p = subp.add_parser('rm', help='remove a task')
        p.add_argument('name', metavar='NAME', help='task name')
        p.set_defaults(task_handler=self.handle_remove)

        p = subp.add_parser('env', help='manage environment')
        p.add_argument('--name', '-n', help='task name')
        ssp = p.add_subparsers()
        p = ssp.add_parser('ls')
        p.set_defaults(task_handler=self.handle_env_list)
        p = ssp.add_parser('set')
        p.add_argument('--file', '-f', help='load from file')
        p.add_argument('values', metavar='VALUES', action=StoreKeyValuePair, nargs='*', help='key value pairs')
        p.set_defaults(task_handler=self.handle_env_set)
        p = ssp.add_parser('unset')
        p.add_argument('values', metavar='VALUES', nargs='+', help='keys to remove')
        p.set_defaults(task_handler=self.handle_env_unset)

        p = subp.add_parser('ports', help='manage ports')
        p.add_argument('name', metavar='NAME', help='task name')
        ssp = p.add_subparsers()
        p = ssp.add_parser('ls')
        p.set_defaults(task_handler=self.handle_ports_list)
        p = ssp.add_parser('set')
        p.add_argument('values', metavar='VALUES', action=StorePortPair, nargs='+', help='port mappings')
        p.set_defaults(task_handler=self.handle_ports_set)
        p = ssp.add_parser('unset')
        p.add_argument('values', metavar='VALUES', nargs='+', help='ports')
        p.set_defaults(task_handler=self.handle_ports_unset)

        p = subp.add_parser('volume', help='manage volumes')
        p.add_argument('name', metavar='NAME', help='task name')
        ssp = p.add_subparsers()
        p = ssp.add_parser('add')
        p.add_argument('volume', metavar='VOLUME', help='volume name')
        p.add_argument('destination', metavar='DEST', help='container mount point')
        p.add_argument('--source', '-s', help='source path')
        p.set_defaults(task_handler=self.handle_volume_add)
        p = ssp.add_parser('remove')
        p.add_argument('volume', metavar='VOLUME', help='volume name')
        p.set_defaults(task_handler=self.handle_volume_remove)

        p = subp.add_parser('command', help='set command')
        p.add_argument('name', metavar='NAME', help='task name')
        p.add_argument('command', metavar='COMMAND', nargs='?', help='command to run')
        p.add_argument('--remove', '-r', action='store_true')
        p.set_defaults(task_handler=self.handle_command)

        p = subp.add_parser('prune', help='remove unused task definitions')
        p.set_defaults(task_handler=self.handle_prune)

    def handle_list(self, args):
        self.list(args.name)

    def list(self, name):
        if name:
            task = self.get_task(name)
            ctr_def = self.get_container_definition(task, name)
            print(json.dumps(ctr_def, indent=2))
        else:
            for task in self.iter_task_families(name):
                ii = task.rfind('-') + 1
                print(task[ii:])

    def handle_make(self, args):
        self.make(args.name, args.image, args.cpu, args.memory)

    def make(self, name, image_name, cpu=None, memory=None, logs=True):
        ctx = self.get_context()
        img_uri = self.client.get_module('image').image_name_to_uri(image_name)
        task = self.get_task(name, ctx=ctx, fail=False)
        if task:
            self.error(f'task "{name}" already exists')
        task = {
            'family': self.get_task_family(name),
            'containerDefinitions': []
        }
        ctr_def = {
            'name': name,
            'image': img_uri,
            'memoryReservation': int(memory or 1),
        }
        if logs:
            ctr_def['logConfiguration'] = {
                'logDriver': 'awslogs',
                'options': {
                    'awslogs-group': f'/{ctx["cluster"]}',
                    'awslogs-regions': ctx['region'],
                    'awslogs-stream-prefix': ctx['app']
                }
            }
        if cpu is not None:
            ctr_def['cpu'] = cpu
        task['containerDefinitions'].append(ctr_def)
        self.register_task(task)

    def handle_remove(self, args):
        self.remove(args.name)

    def remove(self, name):

        # Deregister the task definitions.
        ecs_cli = self.get_boto_client('ecs')
        paginator = ecs_cli.get_paginator('list_task_definitions')
        family = self.get_task_family(name)
        task_defs = paginator.paginate(
            familyPrefix=family
        )
        for results in task_defs:
            for arn in results['taskDefinitionArns']:
                ecs_cli.deregister_task_definition(taskDefinition=arn)

        # Also deregister the launch task definitions (prefixed with underbar).
        task_defs = paginator.paginate(
            familyPrefix='_' + family
        )
        for results in task_defs:
            for arn in results['taskDefinitionArns']:
                ecs_cli.deregister_task_definition(taskDefinition=arn)

    def handle_update(self, args):
        self.update(args.name, args.image, args.cpu, args.memory)

    def update(self, name, image_name, cpu=None, memory=None):
        ctx = self.get_context()
        img_uri = self.client.get_module('image').image_name_to_uri(image_name)
        task = self.get_task(name, ctx=ctx)
        ctr_def = self.get_container_definition(task, name)
        ctr_def['image'] = img_uri
        if cpu is not None:
            ctr_def['cpu'] = int(cpu)
        if memory is not None:
            ctr_def['memoryReservation'] = int(memory)
        self.register_task(task)

    def handle_env_list(self, args):
        self.env_list(args.name)

    def env_list(self, name):
        task = self.get_task(name)
        ctr_def = self.get_container_definition(task, name)
        env = env_to_dict(ctr_def['environment'])
        for k in sorted(env.keys()):
            print('%s=%s' % (k, env[k]))

    def handle_env_set(self, args):
        self.env_set(args.name, values=args.values, file=args.file)

    def env_set(self, name, values=None, file=None):
        to_set = {}
        if file is not None:
            with open(file, 'r') as inf:
                for line in inf:
                    line = line.strip()
                    ii = line.find('=')
                    if ii == -1:
                        continue
                    k = line[:ii]
                    v = line[ii + 1:]
                    to_set[k] = v
        to_set.update(values)
        task = self.get_task(name)
        ctr_def = self.get_container_definition(task, name)
        env = env_to_dict(ctr_def['environment'])
        env.update(to_set)
        env = dict_to_env(env)
        ctr_def['environment'] = env
        self.register_task(task)

    def handle_env_unset(self, args):
        self.env_unset(args.name, args.values)

    def env_unset(self, name, keys):
        app_task = self.get_app_task()
        if not name:
            name = '_'
        ctr_def = self.get_container_definition(app_task, name)
        env = env_to_dict(ctr_def['environment'])
        for k in keys:
            try:
                del env[k]
            except KeyError:
                pass
        env = dict_to_env(env)
        ctr_def['environment'] = env
        self.register_task(app_task)

    def handle_ports_list(self, args):
        self.ports_list(args.name)

    def ports_list(self, name):
        task = self.get_task(name)
        ctr_def = self.get_container_definition(task, name)
        ports = ports_to_dict(ctr_def['portMappings'])
        for k, v in ports.items():
            print('%s:%s' % (k, v))

    def handle_ports_set(self, args):
        self.ports_set(args.name, args.values)

    def ports_set(self, name, values):
        task = self.get_task(name)
        ctr_def = self.get_container_definition(task, name)
        ports = ports_to_dict(ctr_def['portMappings'])
        for k, v in values.items():
            ports[int(k)] = int(v)
        ports = dict_to_ports(ports)
        ctr_def['portMappings'] = ports
        self.register_task(task)

    def handle_ports_unset(self, args):
        self.ports_unset(args.name, args.values)

    def ports_unset(self, name, values):
        app_task = self.get_app_task()
        ctr_def = self.get_container_definition(task, name)
        ports = ports_to_dict(ctr_def['portMappings'])
        for p in values:
            try:
                del ports[int(p)]
            except KeyError:
                pass
        ports = dict_to_ports(ports)
        ctr_def['portMappings'] = ports
        self.register_task(app_task)

    def handle_volume_add(self, args):
        self.volume_add(args.name, args.volume, args.destination, args.source)

    def volume_add(self, task_name, vol_name, dst, src=None, read_only=False):
        task = self.get_task(task_name)
        ctr_def = self.get_container_definition(task, task_name)
        volumes = volumes_to_dict(task['volumes'])
        volumes[vol_name] = src
        task['volumes'] = dict_to_volumes(volumes)
        mounts = mounts_to_dict(ctr_def['mountPoints'])
        mounts[vol_name] = {
            'containerPath': dst,
            'readOnly': read_only
        }
        ctr_def['mountPoints'] = dict_to_mounts(mounts)
        self.register_task(task)

    def handle_volume_remove(self, args):
        self.volume_remove(args.name, args.volume)

    def volume_remove(self, ctr_name, vol_name):
        app_task = self.get_app_task()
        ctr_def = self.get_container_definition(app_task, ctr_name)
        volumes = volumes_to_dict(app_task['volumes'])
        try:
            del volumes[vol_name]
        except KeyError:
            pass
        app_task['volumes'] = dict_to_volumes(volumes)
        mounts = mounts_to_dict(ctr_def['mountPoints'])
        try:
            del mounts[vol_name]
        except KeyError:
            pass
        ctr_def['mountPoints'] = dict_to_mounts(mounts)
        self.register_task(app_task)

    def handle_command(self, args):
        self.command(args.name, args.command, args.remove)

    def command(self, name, cmd, remove=False):
        task = self.get_task(name)
        ctr_def = self.get_container_definition(task, name)
        if remove:
            try:
                del ctr_def['command']
            except KeyError:
                pass
        else:
            ctr_def['command'] = cmd.split()
        self.register_task(task)

    def handle_prune(self, args):
        self.prune()

    def prune(self):
        ecs_cli = self.get_boto_client('ecs')
        for fam in self.iter_task_families():
            for prefix in ['', '_']:
                paginator = ecs_cli.get_paginator('list_task_definitions')
                response = paginator.paginate(
                    familyPrefix=prefix + fam,
                    sort='DESC'
                )
                first = True
                for results in response:
                    for arn in results['taskDefinitionArns']:
                        if not first:
                            ecs_cli.deregister_task_definition(
                                taskDefinition=arn
                            )
                        else:
                            first = False

    def get_task_family(self, name, ctx=None):
        if ctx is None:
            ctx = self.get_context()
        return '-'.join(['fuku', ctx['cluster'], ctx['app']] + ([name] if name else []))

    def get_task(self, name, ctx=None, fail=True):
        family = self.get_task_family(name, ctx)
        ecs = self.get_boto_client('ecs')
        try:
            return ecs.describe_task_definition(
                taskDefinition=family
            )['taskDefinition']
        except:
            if fail:
                self.error(f'no task "{name}"')

    def iter_task_families(self, name=None):
        ctx = self.get_context()
        ecs_cli = self.get_boto_client('ecs')
        paginator = ecs_cli.get_paginator('list_task_definition_families')
        prefix = f'fuku-{ctx["cluster"]}-{ctx["app"]}'
        tasks = paginator.paginate(
            familyPrefix=prefix,
            status='ACTIVE'
        )
        for t in tasks:
            for f in t['families']:
                if f != prefix:
                    yield f

    def register_task(self, task):
        ecs = self.get_boto_client('ecs')
        skip = set(['taskDefinitionArn', 'revision', 'status', 'requiresAttributes'])
        ecs.register_task_definition(**dict([
            (k, v) for k, v in task.items() if k not in skip
        ]))

    def get_container_definition(self, task, name, fail=True):
        if name is None:
            name = self.get_context()['app']
        try:
            return list(filter(lambda x: x['name'] == name, task['containerDefinitions']))[0]
        except (KeyError, IndexError):
            if fail:
                self.error(f'container definition "{name}" does not exist')

    def get_my_context(self):
        return {}



from .module import Module


class Redis(Module):
    dependencies = ['task']

    def __init__(self, **kwargs):
        super().__init__('redis', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='redis help')

        p = subp.add_parser('mk', help='make a redis task')
        p.set_defaults(redis_handler=self.handle_make)

        p = subp.add_parser('connect')
        p.add_argument('target', metavar='TARGET', help='target task name')
        p.set_defaults(redis_handler=self.handle_connect)

    def handle_make(self, args):
        self.make()

    def make(self):
        task_mod = self.client.get_module('task')
        task_mod.make('redis', '!redis:alpine', memory=64)

    def handle_connect(self, args):
        self.connect(args.target)

    def connect(self, target):
        task_mod = self.client.get_module('task')
        env = {
            'REDIS_URL': self.get_url()
        }
        task_mod.env_set(target, env)

    def get_url(self):
        ctx = self.get_context()
        return f'redis://{ctx["app"]}-redis:6379'


class EcsRedis(Redis):
    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='redis help')

        p = subp.add_parser('mk', help='make a redis instance')
        p.add_argument('name', metavar='NAME', help='instance name')
        p.set_defaults(redis_handler=self.handle_make)

        p = subp.add_parser('connect')
        p.add_argument('name', metavar='NAME', help='instance name')
        p.add_argument('target', metavar='TARGET', nargs='?', help='target task name')
        p.set_defaults(redis_handler=self.handle_connect)

    def handle_make(self, args):
        self.make(args.name)

    def make(self, name):
        ctx = self.get_context()
        inst_id = self.get_id(name)
        sg_id = self.client.get_module('cluster').get_security_group_id()
        ec_cli = self.get_boto_client('elasticache')
        ec_cli.create_cache_subnet_group(
            CacheSubnetGroupName=inst_id,
            CacheSubnetGroupDescription=f'Subnet group for {inst_id}',
            SubnetIds=[sn.id for sn in self.get_module('cluster').iter_private_subnets(ctx['cluster'])]
        )
        ec_cli.create_cache_cluster(
            CacheClusterId=name,
            NumCacheNodes=1,
            CacheNodeType='cache.t2.micro',
            Engine='redis',
            CacheSubnetGroupName=inst_id,
            SecurityGroupIds=[sg_id],
            Tags=[
                {
                    'Key': 'cluster',
                    'Value': ctx['cluster']
                },
                {
                    'Key': 'app',
                    'Value': ctx['app']
                }
            ]
        )
        waiter = ec_cli.get_waiter('cache_cluster_available')
        waiter.wait(
            CacheClusterId=name
        )

    def handle_connect(self, args):
        self.connect(args.name, args.target)

    def connect(self, name, target):
        task_mod = self.client.get_module('task')
        env = {
            'REDIS_URL': self.get_url(name)
        }
        task_mod.env_set(target, env)

    def get_id(self, name):
        ctx = self.get_context()
        return f'fuku-{ctx["cluster"]}-{ctx["app"]}-{name}'

    def get_url(self, name):
        ec_cli = self.get_boto_client('elasticache')
        ep = ec_cli.describe_cache_clusters(
            CacheClusterId=name,
            ShowCacheNodeInfo=True
        )['CacheClusters'][0]['CacheNodes'][0]['Endpoint']
        return f'redis://{ep["Address"]}:{ep["Port"]}'



from .module import Module


class Image(Module):
    dependencies = ['app']

    def __init__(self, **kwargs):
        super().__init__('image', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='image help')

        p = subp.add_parser('ls', help='list repositories/images')
        p.set_defaults(image_handler=self.handle_list)

        p = subp.add_parser('mk', help='make a repository')
        p.add_argument('name', metavar='NAME', help='repository name')
        p.set_defaults(image_handler=self.handle_make)

        p = subp.add_parser('connect', help='connect to local image')
        p.add_argument('repo', metavar='REPO', help='repository name')
        p.add_argument('local', metavar='LOCAL', nargs='?', help='local image name')
        p.set_defaults(image_handler=self.handle_connect)

        p = subp.add_parser('push', help='push a connected image')
        p.add_argument('repo', metavar='REPO', help='repository name')
        p.set_defaults(image_handler=self.handle_push)

    def handle_list(self, args):
        self.list()

    def list(self):
        for repo in self.iter_repositories():
            print(repo)

    def handle_make(self, args):
        self.make(args.name)

    def make(self, name):
        ctx = self.get_context()
        ecr = self.get_boto_client('ecr')
        if name in list(self.iter_repositories(ecr=ecr, ctx=ctx)):
            self.error('image by that name already exists')
        if name[0] == '/':
            repo = name[1:]
            self.validate(repo)
        else:
            self.validate(name)
            repo = f'{ctx["app"]}-{name}'
        ecr.create_repository(
            repositoryName=repo
        )

    def handle_connect(self, args):
        self.connect(args.repo, args.local)

    def connect(self, repo, local):
        if repo not in list(self.iter_repositories()):
            self.error(f'repository "{repo}" does not exist')
        if repo[0] == '/':
            x = self.store.setdefault('images', {})
        else:
            x = self.store.setdefault('images', {}).setdefault(app, {})
        x = x.setdefault(repo, {})
        if not local:
            local = x.get('local', None)
            if local:
                print(local)
        else:
            x['local'] = local

    def handle_push(self, args):
        self.push(args.repo)

    def push(self, repo):
        ctx = self.get_context()
        ii = repo.find(':')
        if ii >= 0:
            tag = ':' + repo[ii + 1:]
            repo = repo[:ii]
        else:
            tag = ''
        if repo[0] == '/':
            local = self.store.get('images', {}).get(repo, {}).get('local', None)
        else:
            local = self.store.get('images', {}).get(ctx['app'], {}).get(repo, {}).get('local', None)
        if not local:
            self.error('image not connected')
        ecr = self.get_boto_client('ecr')
        uri = self.get_uri(repo, ctx=ctx, ecr=ecr)
        self.run(f'docker tag {local} {uri}{tag}')
        self.login(ctx=ctx)
        self.run(f'docker push {uri}{tag}', capture=False)

    def iter_repositories(self, ecr=None, ctx=None):
        if ctx is None:
            ctx = self.get_context()
        if ecr is None:
            ecr = self.get_boto_client('ecr')
        data = ecr.describe_repositories()
        data = [d['repositoryName'] for d in data['repositories'] if d['repositoryName'] != 'fuku']
        pre = ctx['app'] + '-'
        results = ['/' + d for d in data if '-' not in d]
        results += [d[len(pre):] for d in data if d.startswith(pre)]
        for res in results:
            yield res

    def get_my_context(self):
        return {}

    def get_uri(self, repo, ecr=None, ctx=None):
        if ctx is None:
            ctx = self.get_context()
        if ecr is None:
            ecr = self.get_boto_client('ecr')
        if repo[0] != '/':
            repo = f'{ctx["app"]}-{repo}'
        else:
            repo = repo[1:]
        ii = repo.rfind(':')
        if ii > -1:
            tag = ':' + repo[ii + 1:]
            repo = repo[:ii]
        else:
            tag = ''
        try:
            return ecr.describe_repositories(
                repositoryNames=[repo]
            )['repositories'][0]['repositoryUri'] + tag
        except KeyError:
            self.error('unknown repository')

    def login(self, ctx=None):
        # TODO: Should really be using the `get_authorization_token` thingo.
        if ctx is None:
            ctx = self.get_context()
        data = self.run(f'aws --profile={ctx["profile"]} --region={ctx["region"]} ecr get-login')
        self.run(data)

    # def create_repository(self, repo):
    #     data = self.run(
    #         '$aws ecr create-repository'
    #         ' --repository-name {}'.format(
    #             repo
    #         ),
    #         capture='json'
    #     )
    #     return data['repository']['repositoryUri']

    # def delete_repository(self, repo):
    #     self.run(
    #         '$aws ecr delete-repository'
    #         ' --repository-name {}'.format(
    #             repo
    #         )
    #     )

    # def get_image(self, name):
    #     app = self.client.get_selected('app')
    #     repo = '%s-%s' % (app, name)
    #     data = self.run(
    #         '$aws ecr describe-images'
    #         ' --repository-name {}'
    #         .format(
    #             repo
    #         ),
    #         capture=True
    #     )
    #     return data

    # def unregister(self, args):
    #     ctr = self.get_image(args.name)
    #     self.delete_repository(ctr['full_repo'])
    #     # del self.store['containers'][ctr['app']][args.name]

    # def launch(self, args):
    #     name = args.name
    #     ctr = self.get_container(name)
    #     mach_mod = self.client.get_module('machine')
    #     mach = mach_mod.get_selected()
    #     mach_mod.ssh_run(
    #         mach['name'],
    #         '/root/pull.sh $region {}'.format(ctr['repo_uri'])
    #     )

    def image_name_to_uri(self, name):
        if name[0] != '!':
            img = self.get_uri(name)
        else:
            img = name[1:]
        return img



import os
import re
import json
import stat
import random

import boto3

from .module import Module
from .utils import entity_already_exists, EntityAlreadyExists
from .db import get_rc_path


ARN_PROG = re.compile(r'[^/]*/fuku-(.+)')


class Cluster(Module):
    dependencies = ['configuration']

    def __init__(self, **kwargs):
        super().__init__('cluster', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='cluster help')

        p = subp.add_parser('ls', help='list clusters')
        p.add_argument('name', metavar='NAME', nargs='?', help='cluster name')
        p.set_defaults(cluster_handler=self.handle_list)

        p = subp.add_parser('mk', help='make a cluster')
        p.add_argument('name', metavar='NAME', help='cluster name')
        p.add_argument('--type', '-t', choices=['swarm', 'ecs'], default='ecs', help='cluster type')
        p.set_defaults(cluster_handler=self.handle_make)

        p = subp.add_parser('sl', help='select a cluster')
        p.add_argument('name', metavar='NAME', help='cluster name')
        p.set_defaults(cluster_handler=self.handle_select)

        p = subp.add_parser('up', help='update a cluster')
        p.add_argument('name', metavar='NAME', help='cluster name')
        p.add_argument('--pem', '-p', help='PEM file')
        p.set_defaults(cluster_handler=self.handle_update)

    def handle_list(self, args):
        self.list(args.name)

    def list(self, name):
        for cl in self.iter_clusters():
            print(cl)

    def handle_make(self, args):
        self.make(args.name, args.type)

    def make(self, name, type):
        self.validate(name)
        vpc = self.create_vpc(name, type)
        self.create_subnets(name, vpc)
        self.create_igw(name)
        nat_id = self.create_nat(name)
        self.create_route_tables(name, nat_id)
        ecs = self.get_boto_client('ecs')
        res = ecs.create_cluster(
            clusterName=f'fuku-{name}'
        )
        ec2 = self.get_boto_client('ec2')
        self.create_key_pair(name, ec2=ec2)
        sg_id = self.create_security_group(name, vpc, ec2=ec2)
        self.create_log_group(name)
        self.create_alb(name, vpc_id, sg_id)
        self.select(name)

    def handle_update(self, args):
        self.update(args.name, args.pem)

    def update(self, name, pem):
        if pem:
            self.add_pem(name, pem_fn=pem)

    def handle_select(self, args):
        self.select(args.name)

    def select(self, name):
        if name and name not in list(self.iter_clusters()):
            self.error(f'no cluster "{name}"')
        self.store_set('selected', name)
        self.clear_parent_selections()

    def iter_clusters(self):
        ecs = self.get_boto_client('ecs')
        for cl in ecs.list_clusters()['clusterArns']:
            m = ARN_PROG.match(cl)
            if not m:
                continue
            yield m.group(1)

    def create_key_pair(self, name, ec2=None):
        if ec2 is None:
            ec2 = self.get_boto_client('ec2')
        try:
            with entity_already_exists(hide=False):
                key = ec2.create_key_pair(
                    KeyName=f'fuku-{name}'
                )['KeyMaterial']
        except EntityAlreadyExists:
            print('key-pair already exists, add PEM file manually')
            return
        self.add_pem(name, pem_key=key)

    def add_pem(self, name, pem_key=None, pem_fn=None):
        ctx = self.get_context()
        path = os.path.join(get_rc_path(), name, 'key.pem')
        try:
            os.makedirs(os.path.dirname(path))
        except OSError:
            pass
        if pem_fn:
            with open(pem_fn, 'r') as pem_f:
                pem_key = pem_f.read()
        with open(path, 'w') as keyf:
            keyf.write(pem_key)
        os.chmod(path, stat.S_IRUSR | stat.S_IWUSR)
        self.run(
            'gpg -c {}'.format(path)
        )
        s3 = self.get_boto_client('s3')
        s3.upload_file(f'{path}.gpg', ctx['bucket'], f'fuku/{ctx["cluster"]}/key.pem')

    def create_vpc(self, name, type):
        ec2 = self.get_boto_resource('ec2')
        vpc = self.get_vpc(name, ec2)
        if not vpc:
            ec2_cli = self.get_boto_client('ec2')
            id = ec2_cli.create_vpc(
                CidrBlock='10.0.0.0/16'
            )['Vpc']['VpcId']
            vpc = ec2.Vpc(id)
            vpc.modify_attribute(EnableDnsSupport={'Value': True})
            vpc.create_tags(
                Tags=[
                    {
                        'Key': 'Name',
                        'Value': f'fuku-{name}',
                    },
                    {
                        'Key': 'cluster',
                        'Value': name,
                    },
                    {
                        'Key': 'type',
                        'Value': type
                    }
                ]
            )
        return vpc

    def create_subnets(self, name, vpc=None):
        if not vpc:
            vpc = self.get_vpc(name)
        public_subnet_a = vpc.create_subnet(
            CidrBlock='10.0.4.0/23',
            AvailabilityZone='ap-southeast-2a'
        )
        public_subnet_a.create_tags(
            Tags=[
                {
                    'Key': 'Name',
                    'Value': f'fuku-{name}-public-a',
                },
                {
                    'Key': 'cluster',
                    'Value': name,
                }
            ]
        )
        public_subnet_b = vpc.create_subnet(
            CidrBlock='10.0.6.0/23',
            AvailabilityZone='ap-southeast-2b'
        )
        public_subnet_b.create_tags(
            Tags=[
                {
                    'Key': 'Name',
                    'Value': f'fuku-{name}-public-b',
                },
                {
                    'Key': 'cluster',
                    'Value': name,
                }
            ]
        )
        private_subnet_a = vpc.create_subnet(
            CidrBlock='10.0.0.0/23'
        )
        private_subnet_a.create_tags(
            Tags=[
                {
                    'Key': 'Name',
                    'Value': f'fuku-{name}-private-a',
                },
                {
                    'Key': 'cluster',
                    'Value': name,
                }
            ]
        )
        private_subnet_b = vpc.create_subnet(
            CidrBlock='10.0.2.0/23'
        )
        private_subnet_b.create_tags(
            Tags=[
                {
                    'Key': 'Name',
                    'Value': f'fuku-{name}-private-b',
                },
                {
                    'Key': 'cluster',
                    'Value': name,
                }
            ]
        )

    def create_nat(self, name, eip=None):
        if not eip:
            eip = self.create_eip()
        subnet = self.get_public_subnet(name)
        ec2_cli = self.get_boto_client('ec2')
        nat_id = ec2_cli.create_nat_gateway(
            SubnetId=subnet.id,
            AllocationId=eip
        )['NatGateway']['NatGatewayId']
        waiter = ec2_cli.get_waiter('nat_gateway_available')
        waiter.wait(NatGatewayIds=[nat_id])
        return nat_id

    def create_igw(self, name):
        igw = self.get_igw(name)
        if igw is not None:
            return igw
        vpc = self.get_vpc(name)
        ec2_cli = self.get_boto_client('ec2')
        id = ec2_cli.create_internet_gateway()['InternetGateway']['InternetGatewayId']
        ec2 = self.get_boto_resource('ec2')
        igw = ec2.InternetGateway(id)
        igw.create_tags(
            Tags=[
                {
                    'Key': 'Name',
                    'Value': f'fuku-{name}'
                },
                {
                    'Key': 'cluster',
                    'Value': name,
                }
            ]
        )
        igw.attach_to_vpc(VpcId=vpc.id)
        return igw

    def create_route_tables(self, name, nat_id):
        ec2_cli = self.get_boto_client('ec2')
        ec2 = self.get_boto_resource('ec2')
        vpc = self.get_vpc(name)
        igw = self.get_igw(name)
        id = ec2_cli.create_route_table(VpcId=vpc.id)['RouteTable']['RouteTableId']
        rt = ec2.RouteTable(id)
        rt.create_tags(
            Tags=[
                {
                    'Key': 'Name',
                    'Value': f'fuku-{name}-private'
                },
                {
                    'Key': 'cluster',
                    'Value': name,
                }
            ]
        )
        ec2_cli.create_route(
            RouteTableId=rt.id,
            DestinationCidrBlock='0.0.0.0/0',
            NatGatewayId=nat_id
        )
        for sn in self.iter_private_subnets(name):
            rt.associate_with_subnet(SubnetId=sn.id)
        id = ec2_cli.create_route_table(VpcId=vpc.id)['RouteTable']['RouteTableId']
        rt = ec2.RouteTable(id)
        rt.create_tags(
            Tags=[
                {
                    'Key': 'Name',
                    'Value': f'fuku-{name}-public'
                },
                {
                    'Key': 'cluster',
                    'Value': name,
                }
            ]
        )
        ec2_cli.create_route(
            RouteTableId=rt.id,
            DestinationCidrBlock='0.0.0.0/0',
            NatGatewayId=igw.id
        )
        for sn in self.iter_public_subnets(name):
            rt.associate_with_subnet(SubnetId=sn.id)

    def create_eip(self):
        ec2_cli = self.get_boto_client('ec2')
        return ec2_cli.allocate_address(
            Domain='vpc'
        )['AllocationId']

    def create_alb(self, name, vpc_id, sg_id):
        alb_cli = self.get_boto_client('elbv2')
        alb_name = f'fuku-{name}-0'
        subnets = [sn.id for sn in self.iter_public_subnets(name)]
        alb_arn = alb_cli.create_load_balancer(
            Name=alb_name,
            Subnets=subnets,
            SecurityGroups=[
                sg_id
            ],
            Scheme='internet-facing',
            IpAddressType='ipv4',
            Tags=[
                {
                    'Key': 'Name',
                    'Value': name
                },
                {
                    'Key': 'cluster',
                    'Value': name
                },
                {
                    'Key': 'index',
                    'Value': '0'
                }
            ]
        )['LoadBalancers'][0]['LoadBalancerArn']
        tg_arn = alb_cli.create_target_group(
            Name=f'fuku-{name}-default',
            Protocol='HTTP',
            Port=80,
            VpcId=vpc_id,
            Matcher={
                'HttpCode': '200,301'
            }
        )['TargetGroups'][0]['TargetGroupArn']
        alb_cli.create_listener(
            LoadBalancerArn=alb_arn,
            Protocol='HTTP',
            Port=80,
            DefaultActions=[
                {
                    'Type': 'forward',
                    'TargetGroupArn': tg_arn
                }
            ]
        )

    def get_alb_arn(self, name, index=0, alb_cli=None):
        if alb_cli is None:
            alb_cli = self.get_boto_client('elbv2')
        return alb_cli.describe_load_balancers(
            Names=[f'fuku-{name}-{index}']
        )['LoadBalancers'][0]['LoadBalancerArn']

    def get_igw(self, name):
        ec2 = self.get_boto_resource('ec2')
        igws = ec2.internet_gateways.filter(Filters=[{'Name': 'tag:cluster', 'Values': [name]}])
        for i in igws:
            return i
        return None

    def get_public_subnet(self, name=None, zone=None):
        if name is None:
            ctx = self.get_context()
            name = ctx['cluster']
        zone = zone if zone else random.choice(['a', 'b'])
        ec2 = self.get_boto_resource('ec2')
        subnets = ec2.subnets.filter(Filters=[{'Name': 'tag:Name', 'Values': [f'fuku-{name}-public-{zone}']}])
        for s in subnets:
            return s
        self.error('no private subnets')

    def iter_public_subnets(self, name, ec2=None):
        ec2 = self.get_boto_resource('ec2')
        subnets = ec2.subnets.filter(Filters=[{'Name': 'tag:Name', 'Values': [f'fuku-{name}-public-a']}])
        for s in subnets:
            yield s
        subnets = ec2.subnets.filter(Filters=[{'Name': 'tag:Name', 'Values': [f'fuku-{name}-public-b']}])
        for s in subnets:
            yield s

    def get_private_subnet(self, name=None, zone=None):
        if name is None:
            ctx = self.get_context()
            name = ctx['cluster']
        zone = zone if zone else random.choice(['a', 'b'])
        ec2 = self.get_boto_resource('ec2')
        subnets = ec2.subnets.filter(Filters=[{'Name': 'tag:Name', 'Values': [f'fuku-{name}-private-{zone}']}])
        for s in subnets:
            return s
        self.error('no private subnets')

    def iter_private_subnets(self, name, ec2=None):
        ec2 = self.get_boto_resource('ec2')
        subnets = ec2.subnets.filter(Filters=[{'Name': 'tag:Name', 'Values': [f'fuku-{name}-private-a']}])
        for s in subnets:
            yield s
        subnets = ec2.subnets.filter(Filters=[{'Name': 'tag:Name', 'Values': [f'fuku-{name}-private-b']}])
        for s in subnets:
            yield s

    def get_vpc(self, name, ec2=None):
        if ec2 is None:
            ec2 = self.get_boto_resource('ec2')
        vpcs = ec2.vpcs.filter(Filters=[{'Name': 'tag:cluster', 'Values': [name]}])
        for vpc in vpcs:
            return vpc
        return None

    def create_security_group(self, name, vpc=None, ec2=None):
        if vpc is None:
            vpc = self.get_vpc(name)
        if ec2 is None:
            ec2 = self.get_boto_client('ec2')
        user_id = self.client.get_module('profile').get_user_id()
        sg_id = None
        with entity_already_exists():
            sg_id = ec2.create_security_group(
                GroupName=f'fuku-{name}',
                Description=f'{name} security group',
                VpcId=vpc.id
            )['GroupId']
        if sg_id is None:
            sg_id = self.get_security_group_id(name)
        with entity_already_exists():
            ec2.authorize_security_group_ingress(
                GroupId=sg_id,
                IpProtocol='tcp',
                FromPort=22,
                ToPort=22,
                CidrIp='0.0.0.0/0'
            )
        with entity_already_exists():
            ec2.authorize_security_group_ingress(
                GroupId=sg_id,
                IpProtocol='tcp',
                FromPort=80,
                ToPort=80,
                CidrIp='0.0.0.0/0'
            )
        with entity_already_exists():
            ec2.authorize_security_group_ingress(
                GroupId=sg_id,
                IpProtocol='tcp',
                FromPort=443,
                ToPort=443,
                CidrIp='0.0.0.0/0'
            )
        with entity_already_exists():
            ec2.authorize_security_group_ingress(
                GroupId=sg_id,
                IpProtocol='tcp',
                FromPort=5432,
                ToPort=5432,
                CidrIp='0.0.0.0/0'
            )
        with entity_already_exists():
            ec2.authorize_security_group_ingress(
                GroupId=sg_id,
                IpPermissions=[{
                    'IpProtocol': '-1',
                    'UserIdGroupPairs': [{
                        'UserId': user_id,
                        'GroupId': sg_id
                    }]
                }]
            )
        return id

    def get_security_group_id(self, name=None):
        name = name or self.store_get('selected')
        ec2 = self.get_boto_client('ec2')
        vpc = self.get_vpc(name)
        all_groups = ec2.describe_security_groups(
            Filters=[{'Name': 'vpc-id', 'Values': [vpc.id]}]
        )['SecurityGroups']
        for sg in all_groups:
            if sg['GroupName'] == f'fuku-{name}':
                return sg['GroupId']
        self.error(f'security group for "{name}" does not exist')

    def create_log_group(self, name):
        logs = self.get_boto_client('logs')
        with entity_already_exists():
            logs.create_log_group(
                logGroupName=f'/{name}'
            )

    def get_my_context(self):
        ctx = {}
        ctx['cluster'] = self.get_selected()
        ctx['pem'] = os.path.join(get_rc_path(), ctx['cluster'], 'key.pem')
        return ctx

    def get_selected(self):
        sel = self.store_get('selected')
        if not sel:
            self.error('no cluster currently selected')
        return sel



import os
from distutils.dir_util import copy_tree

from .module import Module


class Container(Module):
    dependencies = ['machine']

    def __init__(self, **kwargs):
        super().__init__('container', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='container help')

        p = subp.add_parser('list', help='list containers')
        p.add_argument('--all', '-a', action='store_true', help='list inactive containers')
        p.set_defaults(container_handler=self.handle_list)

        # addp = subp.add_parser('add', help='add a machine')
        # addp.set_defaults(machine_handler=self.add)

        # p = subp.add_parser('register', help='register a container definition')
        # p.add_argument('name')
        # p.add_argument('image')
        # p.set_defaults(container_handler=self.handle_register)

        # uregp = subp.add_parser('unregister', help='unregister a container')
        # uregp.add_argument('name', help='container name')
        # uregp.set_defaults(container_handler=self.unregister)

        # pushp = subp.add_parser('push', help='push a container')
        # pushp.add_argument('name', help='container name')
        # pushp.set_defaults(container_handler=self.push)

        p = subp.add_parser('run', help='run a container')
        p.add_argument('--name', '-n', help='container name')
        p.add_argument('--task', '-t', help='task name')
        p.add_argument('--restart', '-r', action='store_true')
        p.set_defaults(container_handler=self.handle_run)

        p = subp.add_parser('remove', help='remove a container')
        p.add_argument('name', help='container name')
        p.set_defaults(container_handler=self.handle_remove)

        p = subp.add_parser('stop', help='remove a container')
        p.add_argument('name', help='container name')
        p.set_defaults(container_handler=self.handle_stop)

        p = subp.add_parser('attach', help='attach to a container')
        p.add_argument('name', help='container name')
        p.add_argument('--shell', '-s', default='bash')
        p.set_defaults(container_handler=self.handle_attach)

        p = subp.add_parser('logs', help='get logs')
        p.add_argument('name', help='container name')
        p.set_defaults(container_handler=self.handle_logs)

        # selp = subp.add_parser('select', help='select a machine')
        # selp.add_argument('instance_id', help='machine instance ID')
        # selp.set_defaults(machine_handler=self.select)

        p = subp.add_parser('template', help='manage container templates')
        p.add_argument('--list', '-l', action='store_true')
        p.add_argument('--copy', '-c')
        p.set_defaults(container_handler=self.handle_template)

    def handle_list(self, args):
        mach_mod = self.client.get_module('machine')
        mach = mach_mod.get_selected()
        cmd = 'fuku-agent list'
        if not args.all:
            cmd += ' -r'
        data = mach_mod.ssh_run(
            cmd,
            name=mach,
            capture='json'
        )
        for ctr in data['result']:
            print(ctr)

    def handle_run(self, args):
        self.run(args.name, args.task, args.restart)

    def run(self, name, task, restart=False):
        mach_mod = self.client.get_module('machine')
        mach = mach_mod.get_selected()
        cmd = 'fuku-agent run'
        if name:
            cmd += ' -n {}'.format(name)
        if task:
            cmd += ' -t {}'.format(task)
        if restart:
            cmd += ' -r'
        data = mach_mod.ssh_run(
            cmd,
            name=mach,
            capture='json'
        )
        if data['status'] != 'ok':
            print(data['result'])

    def handle_remove(self, args):
        mach_mod = self.client.get_module('machine')
        mach = mach_mod.get_selected()
        data = mach_mod.ssh_run(
            'fuku-agent remove {} -d'.format(args.name),
            name=mach,
            capture='json'
        )
        if data['status'] != 'ok':
            print(data['result'])

    def handle_stop(self, args):
        mach_mod = self.client.get_module('machine')
        mach = mach_mod.get_selected()
        data = mach_mod.ssh_run(
            'fuku-agent remove {}'.format(args.name),
            name=mach,
            capture='json'
        )
        if data['status'] != 'ok':
            print(data['result'])

    def handle_attach(self, args):
        mach_mod = self.client.get_module('machine')
        mach = mach_mod.get_selected()
        cmd = 'docker exec -i -t {} {}'.format(args.name, args.shell)
        mach_mod.ssh_run(
            cmd,
            name=mach,
            tty=True,
            capture=False
        )

    def handle_logs(self, args):
        mach_mod = self.client.get_module('machine')
        mach = mach_mod.get_selected()
        cmd = 'docker logs {}'.format(args.name)
        mach_mod.ssh_run(
            cmd,
            name=mach,
            tty=True,
            capture=False
        )

    def handle_template(self, args):
        if args.list:
            for tmpl in os.listdir(self.data_path('images')):
                print(tmpl)
        elif args.copy:
            copy_tree(self.data_path('images/%s' % args.copy), './%s' % args.copy)

    def get_image(self, name):
        pass
        # app = self.client.get_selected('app')
        # ctrs = self.store.get('containers', {}).get(app['name'], {})
        # try:
        #     return ctrs[name]
        # except KeyError:
        #     self.error('unknown container')

    def login(self):
        data = self.run('$aws ecr get-login')
        self.run(data)

    def create_repository(self, repo):
        self.login()
        data = self.run(
            '$aws ecr create-repository'
            ' --repository-name {}'.format(
                repo
            ),
            capture='json'
        )
        return data['repository']['repositoryUri']

    def delete_repository(self, repo):
        self.login()
        self.run(
            '$aws ecr delete-repository'
            ' --repository-name {}'.format(
                repo
            )
        )

    def get_image(self, name):
        app = self.client.get_selected('app')
        repo = '%s-%s' % (app, name)
        self.login()
        data = self.run(
            '$aws describe-images'
            ' --repository-name {}'
            .format(
                repo
            )
        )
        print(data)
        return data

    def register(self, args):
        app = self.client.get_selected('app')
        name = args.name
        repo = '%s-%s' % (app, repo)
        self.create_repository(repo)
        # ctrs = self.store.setdefault('containers', {}).setdefault(app['name'], {})
        # ctrs[name] = {
        #     'name': name,
        #     'repo': repo,
        #     'repo_uri': uri,
        #     'full_repo': full_repo,
        #     'app': app['name'],
        # }

    def unregister(self, args):
        ctr = self.get_image(args.name)
        self.delete_repository(ctr['full_repo'])
        # del self.store['containers'][ctr['app']][args.name]

    def list(self, args):
        app = self.client.get_selected('app')
        # ctrs = self.store.get('containers', {}).get(app['name'], {})
        # for ctr in ctrs.keys():
        #     print('{}'.format(ctr))

    def push(self, args):
        ctr = self.get_image(args.name)
        # uri = ctr['repo_uri']
        self.run(
            'docker tag {}:latest {}:latest'.format(
                ctr['name'],
                uri
            )
        )
        self.run('docker push {}:latest'.format(uri))

    def launch(self, args):
        name = args.name
        ctr = self.get_container(name)
        mach_mod = self.client.get_module('machine')
        mach = mach_mod.get_selected()
        mach_mod.ssh_run(
            mach['name'],
            '/root/pull.sh $region {}'.format(ctr['repo_uri'])
        )



import os
import sys
import json
import tempfile
from contextlib import contextmanager
from string import Template

import boto3

from .runner import run


class Module(object):
    dependencies = []

    def __init__(self, name, db=None, client=None):
        self.name = name
        self.db = db
        self.store = self.db.setdefault(self.name, {})
        self.client = client
        self.use_context = True
        self._checks = {}

    def add_arguments(self, parser):
        pass

    def validate(self, name):
        if '-' in name:
            self.error(f'identifiers cannot contain dashes')
        if name == 'fuku':
            self.error('"fuku" is a reserved name')
        if ' ' in name:
            self.error('no whitespace in names allowed')
        if '/' in name:
            self.error('no slashes in names allowed')

    def get_context(self, ctx={}, use_context=True):
        for dep in self.client.iter_dependent_modules(self):
            ctx = dep.get_context(ctx)
        if self.use_context and use_context:
            ctx.update(self.get_my_context())
        return ctx

    def get_module(self, name):
        return self.client.get_module(name)

    def get_selected(self):
        return None

    def error(self, msg):
        print(msg)
        sys.exit()

    def register_check(self, key, call):
        self._checks[key] = call

    def data_path(self, filename=None):
        path = os.path.join(
            os.path.dirname(
                os.path.realpath(__file__)
            ),
            'scripts',
            'data'
        )
        if filename:
            path = os.path.join(path, filename)
        return path

    @contextmanager
    def template_file(self, filename, context={}):
        # with tempfile.NamedTemporaryFile() as outf:
        with open(self.data_path(filename)) as inf:
            data = Template(inf.read()).substitute(context)
        # outf.write(data.encode())
        # outf.flush()
        # yield outf.name
        yield data

    def run(self, cmd, cfg={}, capture='discard', use_self=False, ignore_errors=False,
            env={}):
        # cfg = self.merged_config(cfg, use_self)
        # final = subs(cmd, cfg)
        # print(final)
        env_copy = os.environ.copy()
        env_copy.update(env)
        output = run(
            # final,
            cmd,
            capture=capture not in set([None, '', False]),
            ignore_errors=ignore_errors,
            env=env_copy
        )
        if capture == 'json':
            output = json.loads(output)
        return output

    def clear_parent_selections(self):
        for parent in self.client.iter_parent_modules(self.name):
            try:
                parent.select(None)
            except AttributeError:
                pass

    def check(self, key):
        if key in self._checks:
            return self._checks[key]()
        else:
            for parent in self.client.iter_parent_modules(self.name):
                res = parent.check(key)
                if res is not None:
                    return res

    def entry(self, args):
        handler = getattr(args, '%s_handler' % self.name, None)
        if handler:
            handler(args)

    def save(self):
        return self.store

    def load(self, cache):
        if 'selected' in cache:
            self.store['selected'] = cache['selected']
        else:
            try:
                del self.store['selected']
            except KeyError:
                pass

    def store_set(self, key, value):
        if value:
            self.store[key] = value
        else:
            try:
                del self.store[key]
            except KeyError:
                pass

    def store_get(self, key):
        parts = key.split('.')
        value = self.store
        for p in parts:
            if value is not None:
                value = value.get(p, None)
        return value

    def db_get(self, key):
        parts = key.split('.')
        value = self.store
        for p in parts:
            if value is not None:
                value = value.get(p, None)
        return value

    def setup_boto_session(self, ctx={}):
        ctx = self.get_context(ctx, use_context=False)
        kwargs = {}
        if 'region' in ctx:
            kwargs['region_name'] = ctx['region']
        if 'profile' in ctx:
            kwargs['profile_name'] = ctx['profile']
        boto3.setup_default_session(**kwargs)

    def get_boto_resource(self, resource, ctx={}):
        self.setup_boto_session(ctx)
        return boto3.resource(resource)

    def get_boto_client(self, resource, ctx={}):
        self.setup_boto_session(ctx)
        return boto3.client(resource)

    def puts3(self, key, value):
        ctx = self.get_context()
        s3 = self.get_boto_client('s3')
        s3.put_object(
            Bucket=ctx['bucket'],
            Key=f'fuku/{key}',
            Body=json.dumps(value)
        )

    def gets3(self, key):
        try:
            ctx = self.get_context()
            s3 = self.get_boto_client('s3')
            data = s3.get_object(
                Bucket=ctx['bucket'],
                Key=f'fuku/{key}',
            )['Body'].read().decode()
            return json.loads(data)
        except:
            return None

    def get_my_context(self):
        return {}
